---
jupytext:
  formats: md:myst
  text_representation:
    extension: .md
    format_name: myst
    format_version: 0.13
    jupytext_version: 1.16.4
kernelspec:
  display_name: Python 3 (ipykernel)
  language: python
  name: python3
mystnb:
  execution_mode: cache
---

```{code-cell} ipython3
:tags: [remove-cell, raises-exception]

import pystencils as ps
import numpy as np
import cupy as cp
```

(codegen_reductions)=
# Code Generation for Reductions

In this guide, we demonstrate how reduction kernels can be generated for different platforms and what impact certain
optimization strategies have.
For this, we set up the update rule for a simple dot product kernel:

```{code-cell} ipython3
r = ps.TypedSymbol("r", "double")
x, y = ps.fields(f"x, y: double[3D]", layout="fzyx")

assign_dot_prod = ps.AddReductionAssignment(r, x.center() * y.center())
```

## CPU Platforms

We first consider a base variant for CPUs without employing any optimizations.
The generated code for this variant looks as follows:

```{code-cell} ipython3
cfg = ps.CreateKernelConfig(target=ps.Target.CurrentCPU)
kernel_cpu = ps.create_kernel(assign_dot_prod, cfg)

ps.inspect(kernel_cpu)
```

We want the reduction kernel to be SIMD vectorized and employ shared-memory parallelism using OpenMP.
The supported SIMD instruction sets for reductions are:
* SSE3
* AVX/AVX2
* AVX512

Below you can see that an AVX vectorization was employed by using the target `Target.X86_AVX`.
**Note that reductions require `assume_inner_stride_one` to be enabled.**
This is due to the fact that other inner strides would require masked SIMD operations 
which are not supported yet.

```{code-cell} ipython3
# configure SIMD vectorization
cfg = ps.CreateKernelConfig(
  target=ps.Target.X86_AVX,
)
cfg.cpu.vectorize.enable = True
cfg.cpu.vectorize.assume_inner_stride_one = True

# configure OpenMP parallelization
cfg.cpu.openmp.enable = True
cfg.cpu.openmp.num_threads = 8

kernel_cpu_opt = ps.create_kernel(assign_dot_prod, cfg)

ps.inspect(kernel_cpu_opt)
```

## GPU Platforms

Reductions are currently only supported for CUDA platforms.
Similar to the CPU section, a base variant for GPUs without explicitly employing any optimizations is shown:

```{code-cell} ipython3
    cfg = ps.CreateKernelConfig(target=ps.Target.CUDA)

    kernel_gpu = ps.create_kernel(assign_dot_prod, cfg)

    ps.inspect(kernel_gpu)
```

As evident from the code, the generated kernel employs atomic operations for updating the pointer 
holding the reduction result.
Using the explicit warp-level instructions provided by CUDA allows us to achieve higher performance compared to
only using atomic operations.
To generate kernels with warp-level reductions, the generator expects that CUDA block sizes are divisible by 
the hardware's warp size.
**Similar to the SIMD configuration, we assure the code generator that the configured block size fulfills this
criterion by enabling `assume_warp_aligned_block_size`.**
While the default block sizes provided by the code generator already fulfill this criterion,
we employ a block fitting algorithm to obtain a block size that is also optimized for the kernel's iteration space.

You can find more detailed information about warp size alignment in {ref}`gpu_codegen`.

```{code-cell} ipython3
    cfg = ps.CreateKernelConfig(target=ps.Target.CUDA)
    cfg.gpu.assume_warp_aligned_block_size = True

    kernel_gpu_opt = ps.create_kernel(assign_dot_prod, cfg)
    
    kernel_func = kernel_gpu_opt.compile()
    kernel_func.launch_config.fit_block_size((32, 1, 1))

    ps.inspect(kernel_gpu_opt)
```

:::{admonition} Developers To Do:

- Support for HIP platforms
- Support vectorization using NEON intrinsics
:::

