from abc import ABC, abstractmethod
from typing import Sequence

from ..ast.expressions import PsCall, PsMemAcc, PsConstantExpr

from ..ast import PsAstNode
from ..functions import (
    CFunction,
    MathFunctions,
    NumericLimitsFunctions,
    ReductionFunctions,
    PsMathFunction,
    PsReductionFunction,
)
from ..literals import PsLiteral
from ...reduction_op_mapping import reduction_op_to_expr
from ...sympyextensions import ReductionOp
from ...types import PsIntegerType, PsIeeeFloatType, PsScalarType, PsPointerType

from .platform import Platform
from ..exceptions import MaterializationError

from ..kernelcreation import AstFactory
from ..kernelcreation.iteration_space import (
    IterationSpace,
    FullIterationSpace,
    SparseIterationSpace,
)

from ..constants import PsConstant
from ..ast.structural import PsDeclaration, PsLoop, PsBlock, PsStructuralNode
from ..ast.expressions import (
    PsSymbolExpr,
    PsExpression,
    PsBufferAcc,
    PsLookup,
    PsGe,
    PsLe,
    PsTernary,
    PsLiteralExpr,
)
from ..ast.vector import PsVecMemAcc
from ...types import PsVectorType, PsCustomType


class GenericCpu(Platform):
    """Generic CPU platform.

    The `GenericCPU` platform models the following execution environment:

     - Generic multicore CPU architecture
     - Iteration space represented by a loop nest, kernels are executed as a whole
     - C standard library math functions available (``#include <math.h>`` or ``#include <cmath>``)
    """

    @property
    def required_headers(self) -> set[str]:
        return {"<cmath>", "<limits>"}

    def materialize_iteration_space(
        self, body: PsBlock, ispace: IterationSpace
    ) -> PsBlock:
        if isinstance(ispace, FullIterationSpace):
            return self._create_domain_loops(body, ispace)
        elif isinstance(ispace, SparseIterationSpace):
            return self._create_sparse_loop(body, ispace)
        else:
            raise MaterializationError(f"Unknown type of iteration space: {ispace}")

    def select_function(
        self, call: PsCall
    ) -> PsExpression | tuple[tuple[PsStructuralNode, ...], PsAstNode]:
        call_func = call.function
        assert isinstance(call_func, PsReductionFunction | PsMathFunction)

        func = call_func.func

        if (
            isinstance(call_func, PsReductionFunction)
            and func is ReductionFunctions.WriteBackToPtr
        ):
            ptr_expr, symbol_expr = call.args
            op = call_func.reduction_op

            assert isinstance(ptr_expr, PsSymbolExpr) and isinstance(
                ptr_expr.dtype, PsPointerType
            )
            assert isinstance(symbol_expr, PsSymbolExpr) and isinstance(
                symbol_expr.dtype, PsScalarType
            )

            ptr_access = PsMemAcc(
                ptr_expr, PsConstantExpr(PsConstant(0, self._ctx.index_dtype))
            )

            # inspired by OpenMP: local reduction variable (negative sign) is added at the end
            actual_op = ReductionOp.Add if op is ReductionOp.Sub else op

            # create binop and potentially select corresponding function for e.g. min or max
            potential_call = reduction_op_to_expr(actual_op, ptr_access, symbol_expr)
            if isinstance(potential_call, PsCall):
                potential_call.dtype = symbol_expr.dtype
                return self.select_function(potential_call)

            return potential_call

        dtype = call.get_dtype()
        arg_types = (dtype,) * func.num_args

        if isinstance(dtype, PsScalarType) and func in (
            NumericLimitsFunctions.Min,
            NumericLimitsFunctions.Max,
        ):
            return PsLiteralExpr(
                PsLiteral(
                    f"std::numeric_limits<{dtype.c_string()}>::{func.function_name}()",
                    dtype,
                )
            )

        if isinstance(dtype, PsIeeeFloatType) and dtype.width in (32, 64):
            cfunc: CFunction
            match func:
                case (
                    MathFunctions.Exp
                    | MathFunctions.Log
                    | MathFunctions.Sin
                    | MathFunctions.Cos
                    | MathFunctions.Tan
                    | MathFunctions.Sinh
                    | MathFunctions.Cosh
                    | MathFunctions.ASin
                    | MathFunctions.ACos
                    | MathFunctions.ATan
                    | MathFunctions.ATan2
                    | MathFunctions.Pow
                    | MathFunctions.Sqrt
                    | MathFunctions.Floor
                    | MathFunctions.Ceil
                ):
                    cfunc = CFunction(func.function_name, arg_types, dtype)
                case MathFunctions.Abs | MathFunctions.Min | MathFunctions.Max:
                    cfunc = CFunction("f" + func.function_name, arg_types, dtype)

            call.function = cfunc
            return call

        if isinstance(dtype, PsIntegerType):
            match func:
                case MathFunctions.Abs:
                    zero = PsExpression.make(PsConstant(0, dtype))
                    arg = call.args[0]
                    return PsTernary(PsGe(arg, zero), arg, -arg)
                case MathFunctions.Min:
                    arg1, arg2 = call.args
                    return PsTernary(PsLe(arg1, arg2), arg1, arg2)
                case MathFunctions.Max:
                    arg1, arg2 = call.args
                    return PsTernary(PsGe(arg1, arg2), arg1, arg2)

        raise MaterializationError(
            f"No implementation available for function {func} on data type {dtype}"
        )

    #   Internals

    def _create_domain_loops(
        self, body: PsBlock, ispace: FullIterationSpace
    ) -> PsBlock:
        factory = AstFactory(self._ctx)

        #   Determine loop order by permuting dimensions
        archetype_field = ispace.archetype_field
        if archetype_field is not None:
            loop_order = archetype_field.layout
        else:
            loop_order = None

        loops = factory.loops_from_ispace(ispace, body, loop_order)
        return PsBlock([loops])

    def _create_sparse_loop(self, body: PsBlock, ispace: SparseIterationSpace):
        factory = AstFactory(self._ctx)

        mappings = [
            PsDeclaration(
                PsSymbolExpr(ctr),
                PsLookup(
                    PsBufferAcc(
                        ispace.index_list.base_pointer,
                        (
                            PsExpression.make(ispace.sparse_counter),
                            factory.parse_index(0),
                        ),
                    ),
                    coord.name,
                ),
            )
            for ctr, coord in zip(ispace.spatial_indices, ispace.coordinate_members)
        ]

        body = PsBlock(mappings + body.statements)

        loop = PsLoop(
            PsSymbolExpr(ispace.sparse_counter),
            PsExpression.make(PsConstant(0, self._ctx.index_dtype)),
            PsExpression.make(ispace.index_list.shape[0]),
            PsExpression.make(PsConstant(1, self._ctx.index_dtype)),
            body,
        )

        return PsBlock([loop])


class GenericVectorCpu(GenericCpu, ABC):
    """Base class for CPU platforms with vectorization support through intrinsics."""

    @abstractmethod
    def type_intrinsic(self, vector_type: PsVectorType) -> PsCustomType:
        """Return the intrinsic vector type for the given generic vector type,
        or raise a `MaterializationError` if type is not supported."""

    @abstractmethod
    def constant_intrinsic(self, c: PsConstant) -> PsExpression:
        """Return an expression that initializes a constant vector,
        or raise a `MaterializationError` if not supported."""

    @abstractmethod
    def op_intrinsic(
        self, expr: PsExpression, operands: Sequence[PsExpression]
    ) -> PsExpression:
        """Return an expression intrinsically invoking the given operation
        or raise a `MaterializationError` if not supported."""

    @abstractmethod
    def math_func_intrinsic(
        self, expr: PsCall, operands: Sequence[PsExpression]
    ) -> PsExpression:
        """Return an expression intrinsically invoking the given mathematical
        function or raise a `MaterializationError` if not supported."""

    @abstractmethod
    def vector_load(self, acc: PsVecMemAcc) -> PsExpression:
        """Return an expression intrinsically performing a vector load,
        or raise a `MaterializationError` if not supported."""

    @abstractmethod
    def vector_store(self, acc: PsVecMemAcc, arg: PsExpression) -> PsExpression:
        """Return an expression intrinsically performing a vector store,
        or raise a `MaterializationError` if not supported."""
