from ..ast.structural import PsAssignment, PsBlock, PsStructuralNode
from ..exceptions import MaterializationError
from ..platforms import Platform
from ..ast import PsAstNode
from ..ast.expressions import PsCall, PsExpression
from ..functions import PsMathFunction, PsReductionFunction


class SelectFunctions:
    """Traverse the AST to replace all instances of `PsMathFunction` by their implementation
    provided by the given `Platform`."""

    def __init__(self, platform: Platform):
        self._platform = platform

    def __call__(self, node: PsAstNode) -> PsAstNode:
        return self.visit(node)

    def visit(self, node: PsAstNode) -> PsAstNode:
        node.children = [self.visit(c) for c in node.children]

        if isinstance(node, PsAssignment):
            rhs = node.rhs
            if isinstance(rhs, PsCall) and isinstance(
                rhs.function, PsReductionFunction
            ):
                resolved_func = self._platform.select_function(rhs)

                match resolved_func:
                    case (prepend, new_rhs):
                        assert isinstance(prepend, tuple)

                        match new_rhs:
                            case PsExpression():
                                return PsBlock(
                                    prepend + (PsAssignment(node.lhs, new_rhs),)
                                )
                            case PsStructuralNode():
                                # special case: produces structural with atomic operation writing value back to ptr
                                return PsBlock(prepend + (new_rhs,))
                            case _:
                                assert False, "Unexpected output from SelectFunctions."
                    case PsExpression():
                        return PsAssignment(node.lhs, resolved_func)
                    case _:
                        raise MaterializationError(
                            f"Wrong return type for resolved function {rhs.function.name} in SelectFunctions."
                        )
            else:
                return node
        elif isinstance(node, PsCall) and isinstance(node.function, PsMathFunction):
            resolved_func = self._platform.select_function(node)
            assert isinstance(resolved_func, PsExpression)

            return resolved_func
        else:
            return node
