import pytest
import numpy as np

import pystencils as ps
from pystencils.sympyextensions import reduction_assignment_from_str

INIT_W = 5
INIT_ARR = 2
SIZE = 15
SOLUTION = {
    "+": INIT_W + INIT_ARR * SIZE,
    "-": INIT_W - INIT_ARR * SIZE,
    "*": INIT_W * INIT_ARR**SIZE,
    "min": min(INIT_W, INIT_ARR),
    "max": max(INIT_W, INIT_ARR),
}


# get AST for kernel with reduction assignment
def get_reduction_assign_ast(dtype, op, config):
    x = ps.fields(f"x: {dtype}[1d]")
    w = ps.TypedSymbol("w", dtype)

    red_assign = reduction_assignment_from_str(w, op, x.center())

    return ps.create_kernel([red_assign], config, default_dtype=dtype)


@pytest.mark.parametrize("instruction_set", ["sse", "avx"])
@pytest.mark.parametrize("dtype", ["float64", "float32"])
@pytest.mark.parametrize("op", ["+", "-", "*", "min", "max"])
def test_reduction_cpu(instruction_set, dtype, op):
    vectorize_info = {
        "instruction_set": instruction_set,
        "assume_inner_stride_one": True,
    }

    config = ps.CreateKernelConfig(
        target=ps.Target.CPU, cpu_openmp=True, cpu_vectorize_info=vectorize_info
    )

    ast_reduction = get_reduction_assign_ast(dtype, op, config)
    ps.show_code(ast_reduction)
    kernel_reduction = ast_reduction.compile()

    array = np.full((SIZE,), INIT_ARR, dtype=dtype)
    reduction_array = np.full((1,), INIT_W, dtype=dtype)

    kernel_reduction(x=array, w=reduction_array)
    assert np.allclose(reduction_array, SOLUTION[op])


@pytest.mark.parametrize("dtype", ["float64", "float32"])
@pytest.mark.parametrize("op", ["+", "-", "*", "min", "max"])
@pytest.mark.parametrize("assume_warp_aligned_block_size", [True, False])
@pytest.mark.parametrize("use_block_fitting", [True, False])
def test_reduction_gpu(
        dtype: str,
        op: str,
        assume_warp_aligned_block_size: bool,
        use_block_fitting: bool,
):
    try:
        import cupy as cp
        from cupy_backends.cuda.api.runtime import CUDARuntimeError

        device_count = range(cp.cuda.runtime.getDeviceCount())
        print(f"Found {device_count} GPUs")
    except ImportError:
        pytest.skip(reason="CuPy is not available", allow_module_level=True)
    except CUDARuntimeError:
        pytest.skip(
            reason="No CUDA capable device is detected", allow_module_level=True
        )

    cfg = ps.CreateKernelConfig(target=ps.Target.GPU)
    cfg.gpu.assume_warp_aligned_block_size = assume_warp_aligned_block_size

    ast_reduction = get_reduction_assign_ast(dtype, op, cfg)
    ps.show_code(ast_reduction)
    kernel_reduction = ast_reduction.compile()

    if use_block_fitting:
        kernel_reduction.launch_config.fit_block_size((32, 1, 1))

    array = np.full((SIZE,), INIT_ARR, dtype=dtype)
    reduction_array = np.full((1,), INIT_W, dtype=dtype)

    array_gpu = cp.asarray(array)
    reduction_array_gpu = cp.asarray(reduction_array)

    kernel_reduction(x=array_gpu, w=reduction_array_gpu)
    assert np.allclose(reduction_array_gpu.get(), SOLUTION[op])
